#include "console-irc.h"
#include "classes-irc.h"
#include "functions-irc.h"
#include "classes-xs-macros.h"
#include "utils.h"

/****** Irssi::Irc ******/
MAKE_ABSTRACT_CLASS(irssi_irc, Irssi::Irc, AC_NONE);
FUN_SPEC_TBL(irssi_irc_proto) = {
  { NULL }
};
FUN_SPEC_TBL(irssi_irc_static) = {
  FUN_DEF(irssi_irc, dcc_chat_find_id, 1), /* string $id */
  FUN_DEF(irssi_irc, dcc_chat_send, 2), /* Irssi::Irc::Dcc::Chat $dcc, string $data */
  FUN_DEF(irssi_irc, dcc_ctcp_message, 5), /* Irssi::Irc::Server $server, string $target, Irssi::Irc::Dcc::Chat $chat, int $notice, string $msg */
  FUN_DEF(irssi_irc, dcc_find_request, 3), /* int $type, string $nick, string $arg */
  FUN_DEF(irssi_irc, dcc_find_request_latest, 1), /* int $type */
  FUN_DEF(irssi_irc, dcc_get_download_path, 1), /* string $fname */
  FUN_DEF(irssi_irc, dcc_register_type, 1), /* string $type */
  FUN_DEF(irssi_irc, dcc_str2type, 1), /* string $str */
  FUN_DEF(irssi_irc, dcc_type2str, 1), /* int $type */
  FUN_DEF(irssi_irc, dcc_unregister_type, 1), /* string $type */
  FUN_DEF(irssi_irc, dccs, 0), /*  */
  FUN_DEF(irssi_irc, get_mask, 3), /* string $nick, string $address, int $flags */
  FUN_DEF(irssi_irc, modes_join, 4), /* Irssi::Irc::Server $server, string $old, string $mode, int $channel */
  FUN_DEF(irssi_irc, notifies, 0), /*  */
  FUN_DEF(irssi_irc, notifylist_add, 4), /* string $mask, string $ircnets, int $away_check, int $idle_check_time */
  FUN_DEF(irssi_irc, notifylist_find, 2), /* string $mask, string $ircnet */
  FUN_DEF(irssi_irc, notifylist_ison, 2), /* string $nick, string $serverlist */
  FUN_DEF(irssi_irc, notifylist_remove, 1), /* string $mask */
  { NULL }
};
SCV_INT_IMPL(irssi_irc, IRC, MASK_DOMAIN)
SCV_INT_IMPL(irssi_irc, IRC, MASK_HOST)
SCV_INT_IMPL(irssi_irc, IRC, MASK_NICK)
SCV_INT_IMPL(irssi_irc, IRC, MASK_USER)
PROP_SPEC_TBL(irssi_irc_static) = {
  PROP_GETTER_DEF(irssi_irc, MASK_DOMAIN),
  PROP_GETTER_DEF(irssi_irc, MASK_HOST),
  PROP_GETTER_DEF(irssi_irc, MASK_NICK),
  PROP_GETTER_DEF(irssi_irc, MASK_USER),
  { NULL }
};

/******** Irssi::Irc::Connect ********/
MAKE_ABSTRACT_CLASS(irssi_irc_connect, Irssi::Irc::Connect, AC_PRV_STORAGE);
#define INNER_IRC_CONNREC_IMPL(klass, nklass, inner_ptr) \
  SG_IMPL(klass, alternate_nick, nklass, STR, inner_ptr->alternate_nick)

#define INNER_IRC_CONNREC_P_G_DEF(klass)			\
    PROP_GETTER_DEF(klass, alternate_nick)

INNER_IRC_CONNREC_IMPL(irssi_irc_connect, IRC_SERVER_CONNECT_REC, N)

PROP_SPEC_TBL(irssi_irc_connect) = {
  INNER_IRC_CONNREC_P_G_DEF(irssi_irc_connect),
  { NULL }
};
FUN_SPEC_TBL(irssi_irc_connect_proto) = { { NULL } };

/****** Irssi::Irc::Server ******/
MAKE_ABSTRACT_CLASS(irssi_irc_server, Irssi::Irc::Server, AC_PRV_STORAGE);

INNER_IRC_CONNREC_IMPL(irssi_irc_server, IRC_SERVER_REC, N-> connrec)

SGS_IMPL(irssi_irc_server, real_address, IRC_SERVER_REC, STR)
SGS_IMPL(irssi_irc_server, usermode, IRC_SERVER_REC, STR)
SGS_IMPL(irssi_irc_server, userhost, IRC_SERVER_REC, STR)

SGS_IMPL(irssi_irc_server, max_cmds_at_once, IRC_SERVER_REC, INT)
SGS_IMPL(irssi_irc_server, cmd_queue_speed, IRC_SERVER_REC, INT)
SGS_IMPL(irssi_irc_server, max_query_chans, IRC_SERVER_REC, INT)

SGS_IMPL(irssi_irc_server, max_kicks_in_cmd, IRC_SERVER_REC, INT)
SGS_IMPL(irssi_irc_server, max_msgs_in_cmd, IRC_SERVER_REC, INT)
SGS_IMPL(irssi_irc_server, max_modes_in_cmd, IRC_SERVER_REC, INT)
SGS_IMPL(irssi_irc_server, max_whois_in_cmd, IRC_SERVER_REC, INT)
SGS_IMPL(irssi_irc_server, isupport_sent, IRC_SERVER_REC, BOOL)

PROP_SPEC_TBL(irssi_irc_server) = {
  INNER_IRC_CONNREC_P_G_DEF(irssi_irc_server),

  PROP_GETTER_DEF(irssi_irc_server, real_address),
  PROP_GETTER_DEF(irssi_irc_server, usermode),
  PROP_GETTER_DEF(irssi_irc_server, userhost),

  PROP_GETTER_DEF(irssi_irc_server, max_cmds_at_once),
  PROP_GETTER_DEF(irssi_irc_server, cmd_queue_speed),
  PROP_GETTER_DEF(irssi_irc_server, max_query_chans),

  PROP_GETTER_DEF(irssi_irc_server, max_kicks_in_cmd),
  PROP_GETTER_DEF(irssi_irc_server, max_msgs_in_cmd),
  PROP_GETTER_DEF(irssi_irc_server, max_modes_in_cmd),
  PROP_GETTER_DEF(irssi_irc_server, max_whois_in_cmd),
  PROP_GETTER_DEF(irssi_irc_server, isupport_sent),
  { NULL }
};

FUN_SPEC_TBL(irssi_irc_server_proto) = {
  FUN_DEF(irssi_irc_server, ctcp_send_reply, 1), /* Irssi::Irc::Server $server, string $data */
  FUN_DEF(irssi_irc_server, get_channels, 0), /* Irssi::Irc::Server $server */
  FUN_DEF(irssi_irc_server, isupport, 1), /* Irssi::Irc::Server $server, string $name */
  FUN_DEF(irssi_irc_server, netsplit_find, 2), /* Irssi::Irc::Server $server, string $nick, string $address */
  FUN_DEF(irssi_irc_server, netsplit_find_channel, 4), /* Irssi::Irc::Server $server, string $nick, string $address, string $channel */
  FUN_DEF(irssi_irc_server, notifylist_ison_server, 1), /* Irssi::Irc::Server $server, string $nick */
  FUN_DEF(irssi_irc_server, redirect_event, 6), /* Irssi::Irc::Server $server, string $command, int $count, string $arg, int $remote, string $failure_signal, SVptr $signals */
  FUN_DEF(irssi_irc_server, redirect_get_signal, 3), /* Irssi::Irc::Server $server, string $prefix, string $event, string $args */
  FUN_DEF(irssi_irc_server, redirect_peek_signal, 3), /* Irssi::Irc::Server $server, string $prefix, string $event, string $args */
  FUN_DEF(irssi_irc_server, send_raw, 1), /* Irssi::Irc::Server $server, string $cmd */
  FUN_DEF(irssi_irc_server, send_raw_first, 1), /* Irssi::Irc::Server $server, string $cmd */
  FUN_DEF(irssi_irc_server, send_raw_now, 1), /* Irssi::Irc::Server $server, string $cmd */
  FUN_DEF(irssi_irc_server, send_raw_split, 3), /* Irssi::Irc::Server $server, string $cmd, int $nickarg, int $max_nicks */
  { NULL }
};
FUN_SPEC_TBL(irssi_irc_server_static) = {
  FUN_DEF(irssi_irc_server, query_create, 3), /* string $server_tag, string $nick, int $automatic */
  FUN_DEF(irssi_irc_server, redirect_register, 5), /* string $command, int $remote, int $timeout, SVptr $start, SVptr $stop, SVptr $opt */
  { NULL }
};

/******** Irssi::Irc::Ban ********/
MAKE_ABSTRACT_CLASS(irssi_irc_ban, Irssi::Irc::Ban, AC_PRV_STORAGE);

SGS_IMPL(irssi_irc_ban, ban, BAN_REC, STR)
SGS_IMPL(irssi_irc_ban, setby, BAN_REC, STR)
SGS_IMPL(irssi_irc_ban, time, BAN_REC, INT)

PROP_SPEC_TBL(irssi_irc_ban) = {
  PROP_GETTER_DEF(irssi_irc_ban, ban),
  PROP_GETTER_DEF(irssi_irc_ban, setby),
  PROP_GETTER_DEF(irssi_irc_ban, time),
  { NULL }
};
FUN_SPEC_TBL(irssi_irc_ban_proto) = { { NULL } };

/******** Irssi::Irc::Channel ********/
MAKE_ABSTRACT_CLASS(irssi_irc_channel, Irssi::Irc::Channel, AC_PRV_STORAGE);

FUN_SPEC_TBL(irssi_irc_channel_proto) = {
  FUN_DEF(irssi_irc_channel, ban_get_mask, 2), /* Irssi::Irc::Channel $channel, string $nick, int $ban_type */
  FUN_DEF(irssi_irc_channel, banlist_add, 3), /* Irssi::Irc::Channel $channel, string $ban, string $nick, time_t $time */
  FUN_DEF(irssi_irc_channel, banlist_remove, 2), /* Irssi::Irc::Channel $channel, string $ban, string $nick */
  FUN_DEF(irssi_irc_channel, bans, 0), /* Irssi::Irc::Channel $channel */
  FUN_DEF(irssi_irc_channel, nick_insert, 5), /* Irssi::Irc::Channel $channel, string $nick, int $op, int $halfop, int $voice, int $send_massjoin */
  { NULL }
};

/******** Irssi::Irc::Dcc ********/
MAKE_ABSTRACT_CLASS(irssi_irc_dcc, Irssi::Irc::Dcc, AC_PRV_STORAGE);

FUN_SPEC_TBL(irssi_irc_dcc_proto) = {
  { NULL }
};
SG_IMPL(irssi_irc_dcc, type, DCC_REC, STR, dcc_type2str(N->type))
SG_IMPL(irssi_irc_dcc, orig_type, DCC_REC, STR, dcc_type2str(N->orig_type))
SGS_IMPL(irssi_irc_dcc, created, DCC_REC, INT)

SGS_IMPL(irssi_irc_dcc, server, DCC_REC, IOBJ)
SGS_IMPL(irssi_irc_dcc, servertag, DCC_REC, STR)
SGS_IMPL(irssi_irc_dcc, mynick, DCC_REC, STR)
SGS_IMPL(irssi_irc_dcc, nick, DCC_REC, STR)

SGS_IMPL(irssi_irc_dcc, chat, DCC_REC, SIOBJ)
SGS_IMPL(irssi_irc_dcc, target, DCC_REC, STR)
SGS_IMPL(irssi_irc_dcc, arg, DCC_REC, STR)

SGS_IMPL(irssi_irc_dcc, addr, DCC_REC, STR)
SGS_IMPL(irssi_irc_dcc, port, DCC_REC, INT)

SGS_IMPL(irssi_irc_dcc, starttime, DCC_REC, INT)
SGS_IMPL(irssi_irc_dcc, transfd, DCC_REC, INT)
PROP_SPEC_TBL(irssi_irc_dcc) = {
  PROP_GETTER_DEF(irssi_irc_dcc, type),
  PROP_GETTER_DEF(irssi_irc_dcc, orig_type),
  PROP_GETTER_DEF(irssi_irc_dcc, created),

  PROP_GETTER_DEF(irssi_irc_dcc, server),
  PROP_GETTER_DEF(irssi_irc_dcc, servertag),
  PROP_GETTER_DEF(irssi_irc_dcc, mynick),
  PROP_GETTER_DEF(irssi_irc_dcc, nick),

  PROP_GETTER_DEF(irssi_irc_dcc, chat),
  PROP_GETTER_DEF(irssi_irc_dcc, target),
  PROP_GETTER_DEF(irssi_irc_dcc, arg),

  PROP_GETTER_DEF(irssi_irc_dcc, addr),
  PROP_GETTER_DEF(irssi_irc_dcc, port),

  PROP_GETTER_DEF(irssi_irc_dcc, starttime),
  PROP_GETTER_DEF(irssi_irc_dcc, transfd),
  { NULL }
};

JSBool
init_classes_irc(JSContext *cx, JSObject *obj) {
  nRV;

  if (!gjs_irssi_cto) {
    g_message(MODULE_NAME ": Irssi class not found");
    return JS_FALSE;
  }
  
  JS_BeginRequest(cx);
  INIT_CLASS_SPROPS(irssi, irc, "Irssi", "Irc");
  INIT_CLASS2_S_PROPS_R(irssi_irc, server, "Irssi_Irc", "Server", Irssi::Server);
  INIT_CLASS2_PROPS_R(irssi_irc, connect, "Irssi_Irc", "Connect", Irssi::Connect);

  INIT_CLASS_PROPS_R(irssi_irc, ban, "Irssi_Irc", "Ban");
  INIT_CLASS2_R(irssi_irc, channel, "Irssi_Irc", "Channel", Irssi::Channel);

  INIT_CLASS_PROPS_R(irssi_irc, dcc, "Irssi_Irc", "Dcc");
  INIT_CLASS2_PROPS_R(irssi_irc_dcc, chat, "Irssi_Irc_Dcc", "Chat", Irssi::Irc::Dcc);
  INIT_CLASS2_PROPS_R(irssi_irc_dcc, file, "Irssi_Irc_Dcc", "File", Irssi::Irc::Dcc);
  INIT_CLASS2_PROPS_R(irssi_irc_dcc, get, "Irssi_Irc_Dcc", "Get", Irssi::Irc::Dcc::File);
  INIT_CLASS2_PROPS_R(irssi_irc_dcc, send, "Irssi_Irc_Dcc", "Send", Irssi::Irc::Dcc::File);
  /*INIT_CLASS_PROPS_R(irssi_irc, netsplit, "Irssi_Irc", "Netsplit");
  INIT_CLASS_PROPS_R(irssi_irc, netsplitserver, "Irssi_Irc", "Netsplitserver");
  INIT_CLASS_PROPS_R(irssi_irc, netsplitchannel, "Irssi_Irc", "Netsplitchannel");
  INIT_CLASS_PROPS_R(irssi_irc, notifylist, "Irssi_Irc", "Notifylist");
  INIT_CLASS_PROPS_R(irssi_irc, client, "Irssi_Irc", "Client");*/
  rv = JS_TRUE;
 out:
  JS_EndRequest(cx);
  return rv;  
}

#if 0
SCRIPT_FUN[Irssi::Irc::Dcc :: close] /* Irssi::Irc::Dcc $dcc */
SCRIPT_FUN[Irssi::Irc::Dcc :: destroy] /* Irssi::Irc::Dcc $dcc */
SCRIPT_FUN[Irssi::Irc::Dcc :: init_rec] /* Irssi::Irc::Dcc $dcc, Irssi::Irc::Server $server, Irssi::Irc::Dcc::Chat $chat, string $nick, string $arg */
SCRIPT_FUN[Irssi::Irc::Dcc :: reject] /* Irssi::Irc::Dcc $dcc, Irssi::Irc::Server $server */
SCRIPT_FUN[Irssi::Irc::Notifylist :: ircnets_match] /* Irssi::Irc::Notifylist $rec, string $ircnet */
#endif
