#include "console.h"
#include "classes.h"
#include "functions.h"
#include "signals.h"
#include "utils.h"
#include "classes-xs-macros.h"

void
gjs_irssi_generic_finalizer(JSFreeOp *fop, JSObject *obj);


FUN_SPEC_TBL(core) = {
  FUN_DEF(core, print, 1),
  { NULL }
};

MAKE_ABSTRACT_CLASS(value_reference, ValueReference, AC_NONE);
FUN_SPEC_TBL(value_reference_proto) = {
  { NULL }
};

/****** Irssi ******/

MAKE_ABSTRACT_CLASS(irssi, Irssi, AC_NONE);
FUN_SPEC_TBL(irssi_proto) = {
  FUN_DEF(irssi, print, 2),
  FUN_DEF(irssi, signal_add, 2),
  FUN_DEF(irssi, signal_remove, 1),
  FUN_DEF(irssi, abstracts_register, 1), /* SVptr $abstracts */
  FUN_DEF(irssi, bits2level, 1), /* int $bits */
  FUN_DEF(irssi, channel_find, 1), /* string $channel */
  FUN_DEF(irssi, chatnet_find, 1), /* string $name */
  FUN_DEF(irssi, combine_level, 2), /* int $level, string $str */
  FUN_DEF(irssi, command, 1), /* string $cmd */
  FUN_DEF(irssi, command_bind, 0), /* ... XXX */
  FUN_DEF(irssi, command_bind_first, 0), /* ... XXX */
  FUN_DEF(irssi, command_bind_last, 0), /* ... XXX */
  FUN_DEF(irssi, command_parse_options, 2), /* string $cmd, string $data */
  FUN_DEF(irssi, command_runsub, 4), /* string $cmd, string $data, Irssi::Server $server, Irssi::Windowitem $item */
  FUN_DEF(irssi, command_set_options, 2), /* string $cmd, string $options */
  FUN_DEF(irssi, command_unbind, 2), /* string $cmd, SVptr $func */
  FUN_DEF(irssi, ctcp_register, 1), /* string $name */
  FUN_DEF(irssi, ctcp_unregister, 1), /* string $name */
  FUN_DEF(irssi, expando_create, 3), /* string $key, SVptr $func, SVptr $signals */
  FUN_DEF(irssi, expando_destroy, 1), /* string $name */
  FUN_DEF(irssi, format_create_dest, 3), /* string $target, int $level, Irssi::UI::Window $window */
  FUN_DEF(irssi, format_get_length, 1), /* string $str */
  FUN_DEF(irssi, format_real_length, 2), /* string $str, int $len */
  FUN_DEF(irssi, get_gui, 0),
  FUN_DEF(irssi, get_irssi_binary, 0),
  FUN_DEF(irssi, get_irssi_config, 0),
  FUN_DEF(irssi, get_irssi_dir, 0),
  FUN_DEF(irssi, gui_input_get_pos, 0),
  FUN_DEF(irssi, gui_input_set, 1), /* string $str */
  FUN_DEF(irssi, gui_input_set_pos, 1), /* int $pos */
  FUN_DEF(irssi, gui_printtext, 3), /* int $xpos, int $ypos, string $str */
  FUN_DEF(irssi, ignore_check, 5), /* string $nick, string $host, string $channel, string $text, int $level */
  FUN_DEF(irssi, input_add, 4), /* int $source, int $condition, SVptr $func, SVptr $data */
  FUN_DEF(irssi, input_remove, 1), /* int $tag */
  FUN_DEF(irssi, level2bits, 1), /* string $str */
  FUN_DEF(irssi, log_create_rec, 2), /* string $fname, int $level */
  FUN_DEF(irssi, log_find, 1), /* string $fname */
  FUN_DEF(irssi, mask_match, 4), /* string $mask, string $nick, string $user, string $host */
  FUN_DEF(irssi, mask_match_address, 3), /* string $mask, string $nick, string $address */
  FUN_DEF(irssi, masks_match, 3), /* string $masks, string $nick, string $address */
  FUN_DEF(irssi, parse_special, 3), /* string $cmd, string $data, int $flags */
  FUN_DEF(irssi, pidwait_add, 1), /* int $pid */
  FUN_DEF(irssi, pidwait_remove, 1), /* int $pid */
  FUN_DEF(irssi, print, 2), /* string $str, int $level */
  FUN_DEF(irssi, printformat, 3), /* int $level, string $format, ... */
  FUN_DEF(irssi, query_find, 1), /* string $nick */
  FUN_DEF(irssi, rawlog_create, 0),
  FUN_DEF(irssi, rawlog_set_size, 1), /* int $lines */
  FUN_DEF(irssi, server_create_conn, 6), /* int $chat_type, string $dest, int $port, string $chatnet, string $password, string $nick */
  FUN_DEF(irssi, server_find_chatnet, 1), /* string $chatnet */
  FUN_DEF(irssi, server_find_tag, 1), /* string $tag */
  FUN_DEF(irssi, settings_add_bool, 3), /* string $section, string $key, int $def */
  FUN_DEF(irssi, settings_add_int, 3), /* string $section, string $key, int $def */
  FUN_DEF(irssi, settings_add_level, 3), /* string $section, string $key, string $def */
  FUN_DEF(irssi, settings_add_size, 3), /* string $section, string $key, string $def */
  FUN_DEF(irssi, settings_add_str, 3), /* string $section, string $key, string $def */
  FUN_DEF(irssi, settings_add_time, 3), /* string $section, string $key, string $def */
  FUN_DEF(irssi, settings_get_bool, 1), /* string $key */
  FUN_DEF(irssi, settings_get_int, 1), /* string $key */
  FUN_DEF(irssi, settings_get_level, 1), /* string $key */
  FUN_DEF(irssi, settings_get_size, 1), /* string $key */
  FUN_DEF(irssi, settings_get_str, 1), /* string $key */
  FUN_DEF(irssi, settings_get_time, 1), /* string $key */
  FUN_DEF(irssi, settings_remove, 1), /* string $key */
  FUN_DEF(irssi, settings_set_bool, 2), /* string $key, int $value */
  FUN_DEF(irssi, settings_set_int, 2), /* string $key, int $value */
  FUN_DEF(irssi, settings_set_level, 2), /* string $key, string $value */
  FUN_DEF(irssi, settings_set_size, 2), /* string $key, string $value */
  FUN_DEF(irssi, settings_set_str, 2), /* string $key, string $value */
  FUN_DEF(irssi, settings_set_time, 2), /* string $key, string $value */
  FUN_DEF(irssi, signal_add, 0), /* ... XXX */
  FUN_DEF(irssi, signal_add_first, 0), /* ... XXX */
  FUN_DEF(irssi, signal_add_last, 0), /* ... XXX */
  FUN_DEF(irssi, signal_add_priority, 0), /* ... XXX */
  FUN_DEF(irssi, signal_continue, 0), /* ... XXX */
  FUN_DEF(irssi, signal_emit, 7), /* string $signal, ... */
  FUN_DEF(irssi, signal_get_emitted, 0),
  FUN_DEF(irssi, signal_get_emitted_id, 0),
  FUN_DEF(irssi, signal_register, 0), /* ... XXX */
  FUN_DEF(irssi, signal_remove, 2), /* string $signal, SVptr $func */
  FUN_DEF(irssi, signal_stop, 0),
  FUN_DEF(irssi, signal_stop_by_name, 1), /* string $signal */
  FUN_DEF(irssi, statusbar_item_register, 3), /* string $name, string $value, string $func */
  FUN_DEF(irssi, statusbar_item_unregister, 1), /* string $name */
  FUN_DEF(irssi, statusbar_items_redraw, 1), /* string $name */
  FUN_DEF(irssi, statusbars_recreate_items, 0),
  FUN_DEF(irssi, strip_codes, 1), /* string $input */
  FUN_DEF(irssi, theme_register, 1), /* SVptr $formats */
  FUN_DEF(irssi, themes_reload, 0),
  FUN_DEF(irssi, timeout_add, 3), /* int $msecs, SVptr $func, SVptr $data */
  FUN_DEF(irssi, timeout_add_once, 3), /* int $msecs, SVptr $func, SVptr $data */
  FUN_DEF(irssi, timeout_remove, 1), /* int $tag */
  FUN_DEF(irssi, version, 0),
  FUN_DEF(irssi, window_find_closest, 2), /* string $name, int $level */
  FUN_DEF(irssi, window_find_item, 1), /* string $name */
  FUN_DEF(irssi, window_find_level, 1), /* int $level */
  FUN_DEF(irssi, window_find_name, 1), /* string $name */
  FUN_DEF(irssi, window_find_refnum, 1), /* int $refnum */
  FUN_DEF(irssi, window_item_find, 1), /* string $name */
  FUN_DEF(irssi, window_refnum_next, 2), /* int $refnum, int $wrap */
  FUN_DEF(irssi, window_refnum_prev, 2), /* int $refnum, int $wrap */
  FUN_DEF(irssi, windows_refnum_last, 0),
  FUN_DEF(irssi, windows, 0),
  FUN_DEF(irssi, chatnets, 0),
  FUN_DEF(irssi, servers, 0),
  FUN_DEF(irssi, queries, 0),
  FUN_DEF(irssi, channels, 0),
  FUN_DEF(irssi, commands, 0),
  FUN_DEF(irssi, ignores, 0),
  FUN_DEF(irssi, reconnects, 0),
  FUN_DEF(irssi, logs, 0),
  FUN_DEF(irssi, term_refresh_freeze, 0), /*  */
  FUN_DEF(irssi, term_refresh_thaw, 0), /*  */
  { NULL }
};

#define INPUT_READ G_INPUT_READ
#define INPUT_WRITE G_INPUT_WRITE

/* Flags for Format expansion */
SC_INT_IMPL(irssi, EXPAND_FLAG_IGNORE_EMPTY)
SC_INT_IMPL(irssi, EXPAND_FLAG_IGNORE_REPLACES)
SC_INT_IMPL(irssi, EXPAND_FLAG_RECURSIVE_MASK)
/* Flags for IO Poll Management */
SC_INT_IMPL(irssi, INPUT_READ)
SC_INT_IMPL(irssi, INPUT_WRITE)
/* Flags for determining the type of frontend irssi is using */
SC_INT_IMPL(irssi, IRSSI_GUI_GNOME)
SC_INT_IMPL(irssi, IRSSI_GUI_GTK)
SC_INT_IMPL(irssi, IRSSI_GUI_KDE)
SC_INT_IMPL(irssi, IRSSI_GUI_NONE)
SC_INT_IMPL(irssi, IRSSI_GUI_QT)
SC_INT_IMPL(irssi, IRSSI_GUI_TEXT)
SC_INT_IMPL(irssi, MSGLEVEL_ACTIONS)
SC_INT_IMPL(irssi, MSGLEVEL_ALL)
SC_INT_IMPL(irssi, MSGLEVEL_CLIENTCRAP)
SC_INT_IMPL(irssi, MSGLEVEL_CLIENTERROR)
SC_INT_IMPL(irssi, MSGLEVEL_CLIENTNOTICE)
SC_INT_IMPL(irssi, MSGLEVEL_CRAP)
SC_INT_IMPL(irssi, MSGLEVEL_CTCPS)
SC_INT_IMPL(irssi, MSGLEVEL_DCC)
SC_INT_IMPL(irssi, MSGLEVEL_DCCMSGS)
SC_INT_IMPL(irssi, MSGLEVEL_HILIGHT)
SC_INT_IMPL(irssi, MSGLEVEL_INVITES)
SC_INT_IMPL(irssi, MSGLEVEL_JOINS)
SC_INT_IMPL(irssi, MSGLEVEL_KICKS)
SC_INT_IMPL(irssi, MSGLEVEL_LASTLOG)
SC_INT_IMPL(irssi, MSGLEVEL_MODES)
SC_INT_IMPL(irssi, MSGLEVEL_MSGS)
SC_INT_IMPL(irssi, MSGLEVEL_NEVER)
SC_INT_IMPL(irssi, MSGLEVEL_NICKS)
SC_INT_IMPL(irssi, MSGLEVEL_NOHILIGHT)
SC_INT_IMPL(irssi, MSGLEVEL_NOTICES)
SC_INT_IMPL(irssi, MSGLEVEL_NO_ACT)
SC_INT_IMPL(irssi, MSGLEVEL_PARTS)
SC_INT_IMPL(irssi, MSGLEVEL_PUBLIC)
SC_INT_IMPL(irssi, MSGLEVEL_QUITS)
SC_INT_IMPL(irssi, MSGLEVEL_SNOTES)
SC_INT_IMPL(irssi, MSGLEVEL_TOPICS)
SC_INT_IMPL(irssi, MSGLEVEL_WALLOPS)
SC_INT_IMPL(irssi, SIGNAL_PRIORITY_DEFAULT)
SC_INT_IMPL(irssi, SIGNAL_PRIORITY_HIGH)
SC_INT_IMPL(irssi, SIGNAL_PRIORITY_LOW)

PROP_SPEC_TBL(irssi_proto) = {

  PROP_GETTER_DEF(irssi, EXPAND_FLAG_IGNORE_EMPTY),
  PROP_GETTER_DEF(irssi, EXPAND_FLAG_IGNORE_REPLACES),
  PROP_GETTER_DEF(irssi, EXPAND_FLAG_RECURSIVE_MASK),

  PROP_GETTER_DEF(irssi, INPUT_READ),
  PROP_GETTER_DEF(irssi, INPUT_WRITE),

  PROP_GETTER_DEF(irssi, IRSSI_GUI_GNOME),
  PROP_GETTER_DEF(irssi, IRSSI_GUI_GTK),
  PROP_GETTER_DEF(irssi, IRSSI_GUI_KDE),
  PROP_GETTER_DEF(irssi, IRSSI_GUI_NONE),
  PROP_GETTER_DEF(irssi, IRSSI_GUI_QT),
  PROP_GETTER_DEF(irssi, IRSSI_GUI_TEXT),

  PROP_GETTER_DEF(irssi, MSGLEVEL_ACTIONS),
  PROP_GETTER_DEF(irssi, MSGLEVEL_ALL),
  PROP_GETTER_DEF(irssi, MSGLEVEL_CLIENTCRAP),
  PROP_GETTER_DEF(irssi, MSGLEVEL_CLIENTERROR),
  PROP_GETTER_DEF(irssi, MSGLEVEL_CLIENTNOTICE),
  PROP_GETTER_DEF(irssi, MSGLEVEL_CRAP),
  PROP_GETTER_DEF(irssi, MSGLEVEL_CTCPS),
  PROP_GETTER_DEF(irssi, MSGLEVEL_DCC),
  PROP_GETTER_DEF(irssi, MSGLEVEL_DCCMSGS),
  PROP_GETTER_DEF(irssi, MSGLEVEL_HILIGHT),
  PROP_GETTER_DEF(irssi, MSGLEVEL_INVITES),
  PROP_GETTER_DEF(irssi, MSGLEVEL_JOINS),
  PROP_GETTER_DEF(irssi, MSGLEVEL_KICKS),
  PROP_GETTER_DEF(irssi, MSGLEVEL_LASTLOG),
  PROP_GETTER_DEF(irssi, MSGLEVEL_MODES),
  PROP_GETTER_DEF(irssi, MSGLEVEL_MSGS),
  PROP_GETTER_DEF(irssi, MSGLEVEL_NEVER),
  PROP_GETTER_DEF(irssi, MSGLEVEL_NICKS),
  PROP_GETTER_DEF(irssi, MSGLEVEL_NOHILIGHT),
  PROP_GETTER_DEF(irssi, MSGLEVEL_NOTICES),
  PROP_GETTER_DEF(irssi, MSGLEVEL_NO_ACT),
  PROP_GETTER_DEF(irssi, MSGLEVEL_PARTS),
  PROP_GETTER_DEF(irssi, MSGLEVEL_PUBLIC),
  PROP_GETTER_DEF(irssi, MSGLEVEL_QUITS),
  PROP_GETTER_DEF(irssi, MSGLEVEL_SNOTES),
  PROP_GETTER_DEF(irssi, MSGLEVEL_TOPICS),
  PROP_GETTER_DEF(irssi, MSGLEVEL_WALLOPS),
  PROP_GETTER_DEF(irssi, SIGNAL_PRIORITY_DEFAULT),
  PROP_GETTER_DEF(irssi, SIGNAL_PRIORITY_HIGH),
  PROP_GETTER_DEF(irssi, SIGNAL_PRIORITY_LOW),
  { NULL }
};

SC_IMPL(irssi, active_win, TOBJ(Irssi::UI::Window), active_win)
SC_IMPL(irssi, active_server, IOBJ, active_win->active_server)
SC_IMPL(irssi, current_theme, TOBJ(Irssi::UI::Theme), current_theme)

PROP_SPEC_TBL(irssi) = {
  PROP_GETTER_DEF(irssi, active_win),
  PROP_GETTER_DEF(irssi, active_server),
  PROP_GETTER_DEF(irssi, current_theme),
  { NULL }
};

/***** CONN_REC ******/

#define INNER_CONNREC_IMPL(klass, nklass, inner_ptr) \
  SG_IMPL(klass, chat_type, nklass, STR, chat_protocol_find_id(inner_ptr->chat_type)->name) \
  									\
  SG_IMPL(klass, address, nklass, STR, inner_ptr->address)		\
  SG_IMPL(klass, port, nklass, INT, inner_ptr->port)			\
  SG_IMPL(klass, chatnet, nklass, STR, inner_ptr->chatnet)		\
  									\
  SG_IMPL(klass, password, nklass, STR, inner_ptr->password)		\
  SG_IMPL(klass, wanted_nick, nklass, STR, inner_ptr->nick)		\
  SG_IMPL(klass, username, nklass, STR, inner_ptr->username)		\
  SG_IMPL(klass, realname, nklass, STR, inner_ptr->realname)		\
  									\
  SG_IMPL(klass, reconnection, nklass, BOOL, inner_ptr->reconnection)	\
  SG_IMPL(klass, no_autojoin_channels, nklass, BOOL, inner_ptr->no_autojoin_channels) \
  SG_IMPL(klass, no_autosendcmd, nklass, BOOL, inner_ptr->no_autosendcmd) \
  SG_IMPL(klass, unix_socket, nklass, INT, inner_ptr->unix_socket)	\
  SG_IMPL(klass, use_ssl, nklass, BOOL, inner_ptr->use_ssl)		\
  SG_IMPL(klass, no_connect, nklass, BOOL, inner_ptr->no_connect)
 
#define INNER_CONNREC_P_G_DEF(klass)			\
    PROP_GETTER_DEF(klass, chat_type),			\
    PROP_GETTER_DEF(klass, address),			\
    PROP_GETTER_DEF(klass, port),			\
    PROP_GETTER_DEF(klass, chatnet),			\
    PROP_GETTER_DEF(klass, password),			\
    PROP_GETTER_DEF(klass, wanted_nick),		\
    PROP_GETTER_DEF(klass, username),			\
    PROP_GETTER_DEF(klass, realname),			\
    PROP_GETTER_DEF(klass, reconnection),		\
    PROP_GETTER_DEF(klass, no_autojoin_channels),	\
    PROP_GETTER_DEF(klass, no_autosendcmd),		\
    PROP_GETTER_DEF(klass, unix_socket),		\
    PROP_GETTER_DEF(klass, use_ssl),			\
    PROP_GETTER_DEF(klass, no_connect)

/****** Irssi::Connect ******/

MAKE_ABSTRACT_CLASS(irssi_connect, Irssi::Connect, AC_PRV_STORAGE);
SG_IMPL(irssi_connect, type, SERVER_CONNECT_REC, STR, "SERVER CONNECT")
INNER_CONNREC_IMPL(irssi_connect, SERVER_CONNECT_REC, N)

PROP_SPEC_TBL(irssi_connect) = {
  PROP_GETTER_DEF(irssi_connect, type),
  INNER_CONNREC_P_G_DEF(irssi_connect),
  { NULL }
};

FUN_SPEC_TBL(irssi_connect_proto) = { { NULL } };

/****** Irssi::Chatnet ******/

MAKE_ABSTRACT_CLASS(irssi_chatnet, Irssi::Chatnet, AC_PRV_STORAGE);

SG_IMPL(irssi_chatnet, type, CHATNET_REC, STR, "CHATNET")
SG_IMPL(irssi_chatnet, chat_type, CHATNET_REC, STR, chat_protocol_find_id(N->chat_type)->name)

SGS_IMPL(irssi_chatnet, name, CHATNET_REC, STR)

SGS_IMPL(irssi_chatnet, nick, CHATNET_REC, STR)
SGS_IMPL(irssi_chatnet, username, CHATNET_REC, STR)
SGS_IMPL(irssi_chatnet, realname, CHATNET_REC, STR)

SGS_IMPL(irssi_chatnet, own_host, CHATNET_REC, STR)
SGS_IMPL(irssi_chatnet, autosendcmd, CHATNET_REC, STR)

PROP_SPEC_TBL(irssi_chatnet) = {
  PROP_GETTER_DEF(irssi_chatnet, type),
  PROP_GETTER_DEF(irssi_chatnet, chat_type),

  PROP_GETTER_DEF(irssi_chatnet, name),

  PROP_GETTER_DEF(irssi_chatnet, nick),
  PROP_GETTER_DEF(irssi_chatnet, username),
  PROP_GETTER_DEF(irssi_chatnet, realname),

  PROP_GETTER_DEF(irssi_chatnet, own_host),
  PROP_GETTER_DEF(irssi_chatnet, autosendcmd),
  { NULL }
};
FUN_SPEC_TBL(irssi_chatnet_proto) = { { NULL } };

/****** Irssi::Server ******/
MAKE_ABSTRACT_CLASS(irssi_server, Irssi::Server, AC_PRV_STORAGE);

SG_IMPL(irssi_server, type, SERVER_REC, STR, "SERVER")

INNER_CONNREC_IMPL(irssi_server, SERVER_REC, N-> connrec)

SGS_IMPL(irssi_server, connect_time, SERVER_REC, INT)
SGS_IMPL(irssi_server, real_connect_time, SERVER_REC, INT)

SGS_IMPL(irssi_server, tag, SERVER_REC, STR)
SGS_IMPL(irssi_server, nick, SERVER_REC, STR)

SGS_IMPL(irssi_server, connected, SERVER_REC, BOOL)
SGS_IMPL(irssi_server, connection_lost, SERVER_REC, BOOL)

SGS_IMPL(irssi_server, rawlog, SERVER_REC, TOBJ(Irssi::Rawlog))

SGS_IMPL(irssi_server, version, SERVER_REC, STR)
SGS_IMPL(irssi_server, away_reason, SERVER_REC, STR)
SGS_IMPL(irssi_server, last_invite, SERVER_REC, STR)
SGS_IMPL(irssi_server, server_operator, SERVER_REC, BOOL)
SGS_IMPL(irssi_server, usermode_away, SERVER_REC, BOOL)
SGS_IMPL(irssi_server, banned, SERVER_REC, BOOL)

SGS_IMPL(irssi_server, lag, SERVER_REC, INT)

PROP_SPEC_TBL(irssi_server) = {
  PROP_GETTER_DEF(irssi_server, type),
  INNER_CONNREC_P_G_DEF(irssi_server),
  PROP_GETTER_DEF(irssi_server, connect_time),
  PROP_GETTER_DEF(irssi_server, real_connect_time),
  PROP_GETTER_DEF(irssi_server, tag),
  PROP_GETTER_DEF(irssi_server, nick),
  PROP_GETTER_DEF(irssi_server, connected),
  PROP_GETTER_DEF(irssi_server, connection_lost),
  PROP_GETTER_DEF(irssi_server, rawlog),
  PROP_GETTER_DEF(irssi_server, version),
  PROP_GETTER_DEF(irssi_server, away_reason),
  PROP_GETTER_DEF(irssi_server, last_invite),
  PROP_GETTER_DEF(irssi_server, server_operator),
  PROP_GETTER_DEF(irssi_server, usermode_away),
  PROP_GETTER_DEF(irssi_server, banned),
  PROP_GETTER_DEF(irssi_server, lag),
  { NULL }
};

FUN_SPEC_TBL(irssi_server_proto) = {
  FUN_DEF(irssi_server, channel_find, 1), /* Irssi::Server $server, string $name */
  FUN_DEF(irssi_server, channels, 0), /* Irssi::Server $server */
  FUN_DEF(irssi_server, channels_join, 2), /* Irssi::Server $server, string $channels, int $automatic */
  FUN_DEF(irssi_server, command, 1), /* Irssi::Server $server, string $cmd */
  FUN_DEF(irssi_server, disconnect, 0), /* Irssi::Server $server */
  FUN_DEF(irssi_server, format_create_dest, 3), /* Irssi::Server $server, string $target, int $level, Irssi::UI::Window $window */
  FUN_DEF(irssi_server, get_nick_flags, 0), /* Irssi::Server $server */
  FUN_DEF(irssi_server, ignore_check, 5), /* Irssi::Server $server, string $nick, string $host, string $channel, string $text, int $level */
  FUN_DEF(irssi_server, ischannel, 1), /* Irssi::Server $server, string $data */
  FUN_DEF(irssi_server, isnickflag, 1), /* Irssi::Server $server, char $flag */
  FUN_DEF(irssi_server, mask_match, 4), /* Irssi::Server $server, string $mask, string $nick, string $user, string $host */
  FUN_DEF(irssi_server, mask_match_address, 3), /* Irssi::Server $server, string $mask, string $nick, string $address */
  FUN_DEF(irssi_server, masks_match, 3), /* Irssi::Server $server, string $masks, string $nick, string $address */
  FUN_DEF(irssi_server, nicks_get_same, 1), /* Irssi::Server $server, string $nick */
  FUN_DEF(irssi_server, parse_special, 3), /* Irssi::Server $server, string $cmd, string $data, int $flags */
  FUN_DEF(irssi_server, print, 3), /* Irssi::Server $server, string $channel, string $str, int $level */
  FUN_DEF(irssi_server, printformat, 3), /* Irssi::Server $server, string $target, int $level, string $format, ... */
  FUN_DEF(irssi_server, queries, 0), /* Irssi::Server $server */
  FUN_DEF(irssi_server, query_find, 1), /* Irssi::Server $server, string $nick */
  /* FUN_DEF(irssi_server, ref, 0), */ /* Irssi::Server $server */
  FUN_DEF(irssi_server, send_message, 3), /* Irssi::Server $server, string $target, string $msg, int $target_type */
  /* FUN_DEF(irssi_server, unref, 0), */ /* Irssi::Server $server */
  FUN_DEF(irssi_server, window_find_closest, 2), /* Irssi::Server $server, string $name, int $level */
  FUN_DEF(irssi_server, window_find_item, 1), /* Irssi::Server $server, string $name */
  FUN_DEF(irssi_server, window_find_level, 1), /* Irssi::Server $server, int $level */
  FUN_DEF(irssi_server, window_item_find, 1), /* Irssi::Server $server, string $name */
  FUN_DEF(irssi_server, gui_printtext_after, 5),
  { NULL }
};

/****** Irssi::Windowitem ******/
MAKE_ABSTRACT_CLASS(irssi_windowitem, Irssi::Windowitem, AC_PRV_STORAGE);

SG_IMPL(irssi_windowitem, type, WI_ITEM_REC, STR, module_find_id_str("WINDOW ITEM TYPE", N-> type))
SG_IMPL(irssi_windowitem, chat_type, WI_ITEM_REC, STR, chat_protocol_find_id(N-> chat_type)->name)

SGS_IMPL(irssi_windowitem, server, WI_ITEM_REC, IOBJ)
SGS_IMPL(irssi_windowitem, visible_name, WI_ITEM_REC, STR)

SGS_IMPL(irssi_windowitem, createtime, WI_ITEM_REC, INT)
SGS_IMPL(irssi_windowitem, data_level, WI_ITEM_REC, INT)
SGS_IMPL(irssi_windowitem, hilight_color, WI_ITEM_REC, STR)

PROP_SPEC_TBL(irssi_windowitem) = {
  PROP_GETTER_DEF(irssi_windowitem, type),
  PROP_GETTER_DEF(irssi_windowitem, chat_type),

  PROP_GETTER_DEF(irssi_windowitem, server),
  PROP_GETTER_DEF(irssi_windowitem, visible_name),

  PROP_GETTER_DEF(irssi_windowitem, createtime),
  PROP_GETTER_DEF(irssi_windowitem, data_level),
  PROP_GETTER_DEF(irssi_windowitem, hilight_color),
  { NULL }
};

FUN_SPEC_TBL(irssi_windowitem_proto) = {
  FUN_DEF(irssi_windowitem, activity, 2), /* Irssi::Windowitem $item, int $data_level, string $hilight_color */
  FUN_DEF(irssi_windowitem, change_server, 1), /* Irssi::Windowitem $item, Irssi::Server $server */
  FUN_DEF(irssi_windowitem, command, 1), /* Irssi::Windowitem $item, string $cmd */
  FUN_DEF(irssi_windowitem, get_dcc, 0), /* Irssi::Windowitem $item */
  FUN_DEF(irssi_windowitem, is_active, 0), /* Irssi::Windowitem $item */
  FUN_DEF(irssi_windowitem, parse_special, 3), /* Irssi::Windowitem $item, string $cmd, string $data, int $flags */
  FUN_DEF(irssi_windowitem, print, 2), /* Irssi::Windowitem $item, string $str, int $level */
  FUN_DEF(irssi_windowitem, printformat, 2), /* Irssi::Windowitem $item, int $level, string $format, ... */
  FUN_DEF(irssi_windowitem, set_active, 0), /* Irssi::Windowitem $item */
  FUN_DEF(irssi_windowitem, window, 0), /* XXX Irssi::Windowitem $item */
  FUN_DEF(irssi_windowitem, window_create, 1), /* Irssi::Windowitem $item, int $automatic */
  { NULL }
};

/****** Irssi::Channel ******/
MAKE_ABSTRACT_CLASS(irssi_channel, Irssi::Channel, AC_PRV_STORAGE);

SGS_IMPL(irssi_channel, ownnick, CHANNEL_REC, IOBJ)
	
SGS_IMPL(irssi_channel, name, CHANNEL_REC, STR)
SGS_IMPL(irssi_channel, topic, CHANNEL_REC, STR)
SGS_IMPL(irssi_channel, topic_by, CHANNEL_REC, STR)
SGS_IMPL(irssi_channel, topic_time, CHANNEL_REC, INT)

SGS_IMPL(irssi_channel, no_modes, CHANNEL_REC, INT)
SGS_IMPL(irssi_channel, mode, CHANNEL_REC, STR)
SGS_IMPL(irssi_channel, limit, CHANNEL_REC, INT)
SGS_IMPL(irssi_channel, key, CHANNEL_REC, STR)

SGS_IMPL(irssi_channel, chanop, CHANNEL_REC, BOOL)
SGS_IMPL(irssi_channel, names_got, CHANNEL_REC, BOOL)
SGS_IMPL(irssi_channel, wholist, CHANNEL_REC, BOOL)
SGS_IMPL(irssi_channel, synced, CHANNEL_REC, BOOL)

SGS_IMPL(irssi_channel, joined, CHANNEL_REC, BOOL)
SGS_IMPL(irssi_channel, left, CHANNEL_REC, BOOL)
SGS_IMPL(irssi_channel, kicked, CHANNEL_REC, BOOL)

PROP_SPEC_TBL(irssi_channel) = {
  PROP_GETTER_DEF(irssi_channel, ownnick),
	
  PROP_GETTER_DEF(irssi_channel, name),
  PROP_GETTER_DEF(irssi_channel, topic),
  PROP_GETTER_DEF(irssi_channel, topic_by),
  PROP_GETTER_DEF(irssi_channel, topic_time),

  PROP_GETTER_DEF(irssi_channel, no_modes),
  PROP_GETTER_DEF(irssi_channel, mode),
  PROP_GETTER_DEF(irssi_channel, limit),
  PROP_GETTER_DEF(irssi_channel, key),

  PROP_GETTER_DEF(irssi_channel, chanop),
  PROP_GETTER_DEF(irssi_channel, names_got),
  PROP_GETTER_DEF(irssi_channel, wholist),
  PROP_GETTER_DEF(irssi_channel, synced),

  PROP_GETTER_DEF(irssi_channel, joined),
  PROP_GETTER_DEF(irssi_channel, left),
  PROP_GETTER_DEF(irssi_channel, kicked),
  { NULL }
};

FUN_SPEC_TBL(irssi_channel_proto) = {
  FUN_DEF(irssi_channel, destroy, 0), /* Irssi::Channel $channel */
  FUN_DEF(irssi_channel, nick_find, 1), /* Irssi::Channel $channel, string $nick */
  FUN_DEF(irssi_channel, nick_find_mask, 1), /* Irssi::Channel $channel, string $mask */
  FUN_DEF(irssi_channel, nick_insert, 1), /* Irssi::Channel $channel, Irssi::Nick $nick */
  FUN_DEF(irssi_channel, nick_remove, 1), /* Irssi::Channel $channel, Irssi::Nick $nick */
  FUN_DEF(irssi_channel, nicks, 0), /* Irssi::Channel $channel */
  { NULL }
};


/****** Irssi::Query ******/
MAKE_ABSTRACT_CLASS(irssi_query, Irssi::Query, AC_PRV_STORAGE);

SGS_IMPL(irssi_query, name, QUERY_REC, STR)
SGS_IMPL(irssi_query, last_unread_msg, QUERY_REC, INT)
SGS_IMPL(irssi_query, address, QUERY_REC, STR)
SGS_IMPL(irssi_query, server_tag, QUERY_REC, STR)
SGS_IMPL(irssi_query, unwanted, QUERY_REC, BOOL)

PROP_SPEC_TBL(irssi_query) = {
  PROP_GETTER_DEF(irssi_query, name),
  PROP_GETTER_DEF(irssi_query, last_unread_msg),
  PROP_GETTER_DEF(irssi_query, address),
  PROP_GETTER_DEF(irssi_query, server_tag),
  PROP_GETTER_DEF(irssi_query, unwanted),
  { NULL }
};

FUN_SPEC_TBL(irssi_query_proto) = {
  FUN_DEF(irssi_query, change_server, 1), /* Irssi::Query $query, Irssi::Server $server */
  FUN_DEF(irssi_query, destroy, 0),
  { NULL }
};

/****** Irssi::Nick ******/
MAKE_ABSTRACT_CLASS(irssi_nick, Irssi::Nick, AC_PRV_STORAGE);

FUN_SPEC_TBL(irssi_nick_proto) = {
  { NULL }
};

SG_IMPL(irssi_nick, type, NICK_REC, STR, "NICK")
SG_IMPL(irssi_nick, chat_type, NICK_REC, STR, chat_protocol_find_id(N-> chat_type)->name)

SGS_IMPL(irssi_nick, nick, NICK_REC, STR)
SGS_IMPL(irssi_nick, host, NICK_REC, STR)
SGS_IMPL(irssi_nick, realname, NICK_REC, STR)
SGS_IMPL(irssi_nick, hops, NICK_REC, INT)

SGS_IMPL(irssi_nick, gone, NICK_REC, BOOL)
SGS_IMPL(irssi_nick, serverop, NICK_REC, BOOL)

SGS_IMPL(irssi_nick, op, NICK_REC, BOOL)
SGS_IMPL(irssi_nick, halfop, NICK_REC, BOOL)
SGS_IMPL(irssi_nick, voice, NICK_REC, BOOL)
SG_IMPL(irssi_nick, other, NICK_REC, BOOL, N->prefixes[0])
SGS_IMPL(irssi_nick, prefixes, NICK_REC, STR)

SGS_IMPL(irssi_nick, last_check, NICK_REC, INT)
SGS_IMPL(irssi_nick, send_massjoin, NICK_REC, BOOL)

PROP_SPEC_TBL(irssi_nick) = {
  PROP_GETTER_DEF(irssi_nick, type),
  PROP_GETTER_DEF(irssi_nick, chat_type),

  PROP_GETTER_DEF(irssi_nick, nick),
  PROP_GETTER_DEF(irssi_nick, host),
  PROP_GETTER_DEF(irssi_nick, realname),
  PROP_GETTER_DEF(irssi_nick, hops),

  PROP_GETTER_DEF(irssi_nick, gone),
  PROP_GETTER_DEF(irssi_nick, serverop),

  PROP_GETTER_DEF(irssi_nick, op),
  PROP_GETTER_DEF(irssi_nick, halfop),
  PROP_GETTER_DEF(irssi_nick, voice),
  PROP_GETTER_DEF(irssi_nick, other),
  PROP_GETTER_DEF(irssi_nick, prefixes),

  PROP_GETTER_DEF(irssi_nick, last_check),
  PROP_GETTER_DEF(irssi_nick, send_massjoin),
  { NULL }
};

/****** Irssi::Command ******/
MAKE_ABSTRACT_CLASS(irssi_command, Irssi::Command, AC_PRV_STORAGE);
SGS_IMPL(irssi_command, category, COMMAND_REC, STR)
SGS_IMPL(irssi_command, cmd, COMMAND_REC, STR)

PROP_SPEC_TBL(irssi_command) = {
  PROP_GETTER_DEF(irssi_command, category),
  PROP_GETTER_DEF(irssi_command, cmd),
  { NULL }
};
FUN_SPEC_TBL(irssi_command_proto) = { { NULL } };

/****** Irssi::Ignore ******/
MAKE_ABSTRACT_CLASS(irssi_ignore, Irssi::Ignore, AC_PRV_STORAGE);
SGS_IMPL(irssi_ignore, mask, IGNORE_REC, STR)
SGS_IMPL(irssi_ignore, servertag, IGNORE_REC, STR)
SGS_IMPL(irssi_ignore, channels, IGNORE_REC, STRPTRARR)
SGS_IMPL(irssi_ignore, pattern, IGNORE_REC, STR)

SGS_IMPL(irssi_ignore, level, IGNORE_REC, INT)

SGS_IMPL(irssi_ignore, exception, IGNORE_REC, BOOL)
SGS_IMPL(irssi_ignore, regexp, IGNORE_REC, BOOL)
SGS_IMPL(irssi_ignore, fullword, IGNORE_REC, BOOL)

PROP_SPEC_TBL(irssi_ignore) = {
  PROP_GETTER_DEF(irssi_ignore, mask),
  PROP_GETTER_DEF(irssi_ignore, servertag),
  PROP_GETTER_DEF(irssi_ignore, channels),
  PROP_GETTER_DEF(irssi_ignore, pattern),

  PROP_GETTER_DEF(irssi_ignore, level),

  PROP_GETTER_DEF(irssi_ignore, exception),
  PROP_GETTER_DEF(irssi_ignore, regexp),
  PROP_GETTER_DEF(irssi_ignore, fullword),
  { NULL }
};
FUN_SPEC_TBL(irssi_ignore_proto) = { { NULL } };


/****** Irssi::Log ******/
MAKE_ABSTRACT_CLASS(irssi_log, Irssi::Log, AC_PRV_STORAGE);
SGS_IMPL(irssi_log, fname, LOG_REC, STR)
SGS_IMPL(irssi_log, real_fname, LOG_REC, STR)
SGS_IMPL(irssi_log, opened, LOG_REC, BOOL)
SGS_IMPL(irssi_log, level, LOG_REC, INT)
SGS_IMPL(irssi_log, last, LOG_REC, INT)
SGS_IMPL(irssi_log, autoopen, LOG_REC, BOOL)
SGS_IMPL(irssi_log, failed, LOG_REC, BOOL)
SGS_IMPL(irssi_log, temp, LOG_REC, BOOL)

SGS_IMPL(irssi_log, items, LOG_REC, TOBJARR(Irssi::Logitem))

PROP_SPEC_TBL(irssi_log) = {
  PROP_GETTER_DEF(irssi_log, fname),
  PROP_GETTER_DEF(irssi_log, real_fname),
  PROP_GETTER_DEF(irssi_log, opened),
  PROP_GETTER_DEF(irssi_log, level),
  PROP_GETTER_DEF(irssi_log, last),
  PROP_GETTER_DEF(irssi_log, autoopen),
  PROP_GETTER_DEF(irssi_log, failed),
  PROP_GETTER_DEF(irssi_log, temp),

  PROP_GETTER_DEF(irssi_log, items),
  { NULL }
};
FUN_SPEC_TBL(irssi_log_proto) = {
  FUN_DEF(irssi_log, close, 0), /* Irssi::Log $log */
  FUN_DEF(irssi_log, item_add, 3), /* Irssi::Log $log, int $type, string $name, string $servertag */
  FUN_DEF(irssi_log, item_destroy, 1), /* Irssi::Log $log, Irssi::Logitem $item */
  FUN_DEF(irssi_log, item_find, 3), /* Irssi::Log $log, int $type, string $item, string $servertag */
  FUN_DEF(irssi_log, start_logging, 0), /* Irssi::Log $log */
  FUN_DEF(irssi_log, stop_logging, 0), /* Irssi::Log $log */
  FUN_DEF(irssi_log, update, 0), /* Irssi::Log $log */
  FUN_DEF(irssi_log, write_rec, 2), /* Irssi::Log $log, string $str, int $level */
  { NULL }
};

/****** Irssi::Logitem ******/
MAKE_ABSTRACT_CLASS(irssi_logitem, Irssi::Logitem, AC_PRV_STORAGE);
SGS_IMPL(irssi_logitem, type, LOG_ITEM_REC, INT)
SGS_IMPL(irssi_logitem, name, LOG_ITEM_REC, STR)
SGS_IMPL(irssi_logitem, servertag, LOG_ITEM_REC, STR)

PROP_SPEC_TBL(irssi_logitem) = {
  PROP_GETTER_DEF(irssi_logitem, type),
  PROP_GETTER_DEF(irssi_logitem, name),
  PROP_GETTER_DEF(irssi_logitem, servertag),
  { NULL }
};
FUN_SPEC_TBL(irssi_logitem_proto) = { { NULL } };

/****** Irssi::Rawlog ******/
MAKE_ABSTRACT_CLASS(irssi_rawlog, Irssi::Rawlog, AC_PRV_STORAGE);
SGS_IMPL(irssi_rawlog, logging, RAWLOG_REC, BOOL)
SGS_IMPL(irssi_rawlog, nlines, RAWLOG_REC, INT)
PROP_SPEC_TBL(irssi_rawlog) = {
  PROP_GETTER_DEF(irssi_rawlog, logging),
  PROP_GETTER_DEF(irssi_rawlog, nlines),
  { NULL }
};
FUN_SPEC_TBL(irssi_rawlog_proto) = {
  FUN_DEF(irssi_rawlog, close, 0), /* Irssi::Rawlog $rawlog */
  FUN_DEF(irssi_rawlog, destroy, 0), /* Irssi::Rawlog $rawlog */
  FUN_DEF(irssi_rawlog, get_lines, 0), /* Irssi::Rawlog $rawlog */
  FUN_DEF(irssi_rawlog, input, 1), /* Irssi::Rawlog $rawlog, string $str */
  FUN_DEF(irssi_rawlog, open, 1), /* Irssi::Rawlog $rawlog, string $fname */
  FUN_DEF(irssi_rawlog, output, 1), /* Irssi::Rawlog $rawlog, string $str */
  FUN_DEF(irssi_rawlog, redirect, 1), /* Irssi::Rawlog $rawlog, string $str */
  FUN_DEF(irssi_rawlog, save, 1), /* Irssi::Rawlog $rawlog, string $fname */
  { NULL }
};

MAKE_ABSTRACT_CLASS(irssi_reconnect, Irssi::Reconnect, AC_PRV_STORAGE);
SG_IMPL(irssi_reconnect, type, RECONNECT_REC, STR, "RECONNECT")
INNER_CONNREC_IMPL(irssi_reconnect, RECONNECT_REC, N-> conn)
SGS_IMPL(irssi_reconnect, tag, RECONNECT_REC, INT)
SGS_IMPL(irssi_reconnect, next_connect, RECONNECT_REC, INT)

PROP_SPEC_TBL(irssi_reconnect) = {
  PROP_GETTER_DEF(irssi_reconnect, type),
  INNER_CONNREC_P_G_DEF(irssi_reconnect),
  PROP_GETTER_DEF(irssi_reconnect, tag),
  PROP_GETTER_DEF(irssi_reconnect, next_connect),
  { NULL }
};

/****** Irssi::UI ******/
MAKE_ABSTRACT_CLASS(irssi_ui, Irssi::UI, AC_NONE);

FUN_SPEC_TBL(irssi_ui_proto) = {
  { NULL }
};
FUN_SPEC_TBL(irssi_ui_static) = {
  FUN_DEF(irssi_ui, processes, 0),
  { NULL }
};

/****** Irssi::UI::Window *******/
MAKE_ABSTRACT_CLASS(irssi_ui_window, Irssi::UI::Window, AC_PRV_STORAGE);

FUN_SPEC_TBL(irssi_ui_window_proto) = {
  FUN_DEF(irssi_ui_window, format_create_dest, 1),
  FUN_DEF(irssi_ui_window, activity, 2),
  FUN_DEF(irssi_ui_window, change_server, 1),
  FUN_DEF(irssi_ui_window, command, 1),
  FUN_DEF(irssi_ui_window, format_get_text, 4), /* Irssi::UI::Window $window, string $module, Irssi::Server $server, string $target, int $formatnum, ... */
  FUN_DEF(irssi_ui_window, get_history_lines, 0),
  FUN_DEF(irssi_ui_window, gui_printtext_after, 4),
  FUN_DEF(irssi_ui_window, item_add, 2),
  FUN_DEF(irssi_ui_window, item_find, 2),
  FUN_DEF(irssi_ui_window, last_line_insert, 0),
  FUN_DEF(irssi_ui_window, print, 2),
  FUN_DEF(irssi_ui_window, print_after, 4),
  FUN_DEF(irssi_ui_window, printformat, 2), /* Irssi::UI::Window $window, int $level, string $format, ... */
  FUN_DEF(irssi_ui_window, set_name, 1),
  FUN_DEF(irssi_ui_window, destroy, 0),
  FUN_DEF(irssi_ui_window, get_active_name, 0),
  FUN_DEF(irssi_ui_window, item_next, 0),
  FUN_DEF(irssi_ui_window, item_prev, 0),
  FUN_DEF(irssi_ui_window, set_active, 0),
  FUN_DEF(irssi_ui_window, set_history, 1),
  FUN_DEF(irssi_ui_window, set_level, 1),
  FUN_DEF(irssi_ui_window, set_refnum, 1),
  FUN_DEF(irssi_ui_window, view, 0),
  FUN_DEF(irssi_ui_window, items, 0),
  { NULL }
};
FUN_SPEC_TBL(irssi_ui_window_static) = {
  FUN_DEF(irssi_ui_window, item_destroy, 1),
  FUN_DEF(irssi_ui_window, item_remove, 1),
  { NULL }
};
SGS_IMPL(irssi_ui_window, refnum, WINDOW_REC, INT)
SGS_IMPL(irssi_ui_window, name, WINDOW_REC, STR)
SGS_IMPL(irssi_ui_window, history_name, WINDOW_REC, STR)
SGS_IMPL(irssi_ui_window, width, WINDOW_REC, INT)
SGS_IMPL(irssi_ui_window, height, WINDOW_REC, INT)
SGS_IMPL(irssi_ui_window, active, WINDOW_REC, IOBJ)
SGS_IMPL(irssi_ui_window, active_server, WINDOW_REC, IOBJ)
SGS_IMPL(irssi_ui_window, servertag, WINDOW_REC, STR)
SGS_IMPL(irssi_ui_window, level, WINDOW_REC, INT)
SGS_IMPL(irssi_ui_window, immortal, WINDOW_REC, INT)
SGS_IMPL(irssi_ui_window, sticky_refnum, WINDOW_REC, INT)
SGS_IMPL(irssi_ui_window, data_level, WINDOW_REC, INT)
SGS_IMPL(irssi_ui_window, hilight_color, WINDOW_REC, STR)
SGS_IMPL(irssi_ui_window, last_timestamp, WINDOW_REC, INT)
SGS_IMPL(irssi_ui_window, last_line, WINDOW_REC, INT)
SGS_IMPL(irssi_ui_window, theme, WINDOW_REC, TOBJ(Irssi::UI::Theme))
SGS_IMPL(irssi_ui_window, theme_name, WINDOW_REC, STR)

PROP_SPEC_TBL(irssi_ui_window) = {
  PROP_GETTER_DEF(irssi_ui_window, refnum),
  PROP_GETTER_DEF(irssi_ui_window, name),
  PROP_GETTER_DEF(irssi_ui_window, history_name),
  PROP_GETTER_DEF(irssi_ui_window, width),
  PROP_GETTER_DEF(irssi_ui_window, height),
  PROP_GETTER_DEF(irssi_ui_window, active),
  PROP_GETTER_DEF(irssi_ui_window, active_server),
  PROP_GETTER_DEF(irssi_ui_window, servertag),
  PROP_GETTER_DEF(irssi_ui_window, level),
  PROP_GETTER_DEF(irssi_ui_window, immortal),
  PROP_GETTER_DEF(irssi_ui_window, sticky_refnum),
  PROP_GETTER_DEF(irssi_ui_window, data_level),
  PROP_GETTER_DEF(irssi_ui_window, hilight_color),
  PROP_GETTER_DEF(irssi_ui_window, last_timestamp),
  PROP_GETTER_DEF(irssi_ui_window, last_line),
  PROP_GETTER_DEF(irssi_ui_window, theme),
  PROP_GETTER_DEF(irssi_ui_window, theme_name),
  { NULL }
};

/****** Irssi::UI::TextDest *******/
MAKE_ABSTRACT_CLASS(irssi_ui_textdest, Irssi::UI::TextDest, AC_PRV_STORAGE);

FUN_SPEC_TBL(irssi_ui_textdest_proto) = {
  FUN_DEF(irssi_ui_textdest, print, 1), /* Irssi::UI::TextDest $dest, string $str */
  { NULL }
};

SGS_IMPL(irssi_ui_textdest, window, TEXT_DEST_REC, TOBJ(Irssi::UI::Window))
SGS_IMPL(irssi_ui_textdest, server, TEXT_DEST_REC, IOBJ)
SGS_IMPL(irssi_ui_textdest, target, TEXT_DEST_REC, STR)
SGS_IMPL(irssi_ui_textdest, level, TEXT_DEST_REC, INT)
SGS_IMPL(irssi_ui_textdest, hilight_color, TEXT_DEST_REC, STR)
SGS_IMPL(irssi_ui_textdest, hilight_priority, TEXT_DEST_REC, INT)

PROP_SPEC_TBL(irssi_ui_textdest) = {
  PROP_GETTER_DEF(irssi_ui_textdest, window),
  PROP_GETTER_DEF(irssi_ui_textdest, server),
  PROP_GETTER_DEF(irssi_ui_textdest, target),
  PROP_GETTER_DEF(irssi_ui_textdest, level),
  PROP_GETTER_DEF(irssi_ui_textdest, hilight_color),
  PROP_GETTER_DEF(irssi_ui_textdest, hilight_priority),
  { NULL }
};

/****** Irssi::TextUI ******/
MAKE_ABSTRACT_CLASS(irssi_textui, Irssi::TextUI, AC_NONE);

FUN_SPEC_TBL(irssi_textui_proto) = {
  { NULL }
};
/****** Irssi::TextUI::Line ******/
MAKE_ABSTRACT_CLASS(irssi_textui_line, Irssi::TextUI::Line, AC_PRV_STORAGE);
FUN_SPEC_TBL(irssi_textui_line_proto) = {
  FUN_DEF(irssi_textui_line, get_text, 1),
  FUN_DEF(irssi_textui_line, next, 0),
  FUN_DEF(irssi_textui_line, prev, 0),
  { NULL }
};
SG_IMPL(irssi_textui_line, info, LINE_REC, TOBJ(Irssi::TextUI::LineInfo), &(N->info))
PROP_SPEC_TBL(irssi_textui_line) = {
  PROP_GETTER_DEF(irssi_textui_line, info),  
  { NULL }
};
/****** Irssi::TextUI::LineInfo ******/
MAKE_ABSTRACT_CLASS(irssi_textui_lineinfo, Irssi::TextUI::LineInfo, AC_PRV_STORAGE);
FUN_SPEC_TBL(irssi_textui_lineinfo_proto) = {
  { NULL }
};
SGS_IMPL(irssi_textui_lineinfo, level, LINE_INFO_REC, INT)
SGS_IMPL(irssi_textui_lineinfo, time, LINE_INFO_REC, INT)

PROP_SPEC_TBL(irssi_textui_lineinfo) = {
  PROP_GETTER_DEF(irssi_textui_lineinfo, level),  
  PROP_GETTER_DEF(irssi_textui_lineinfo, time),
  { NULL }
};

/****** Irssi::TextUI::LineCache ******/
MAKE_ABSTRACT_CLASS(irssi_textui_linecache, Irssi::TextUI::LineCache, AC_PRV_STORAGE);
FUN_SPEC_TBL(irssi_textui_linecache_proto) = { { NULL } };
SGS_IMPL(irssi_textui_linecache, last_access, LINE_CACHE_REC, INT)
SGS_IMPL(irssi_textui_linecache, count, LINE_CACHE_REC, INT)
PROP_SPEC_TBL(irssi_textui_linecache) = {
  PROP_GETTER_DEF(irssi_textui_linecache, last_access),
  PROP_GETTER_DEF(irssi_textui_linecache, count),
  { NULL }
};

/****** Irssi::TextUI::TextBuffer ******/
MAKE_ABSTRACT_CLASS(irssi_textui_textbuffer, Irssi::TextUI::TextBuffer, AC_PRV_STORAGE);
FUN_SPEC_TBL(irssi_textui_textbuffer_proto) = { { NULL } };
SGS_IMPL(irssi_textui_textbuffer, first_line, TEXT_BUFFER_REC, TOBJ(Irssi::TextUI::Line))
SGS_IMPL(irssi_textui_textbuffer, lines_count, TEXT_BUFFER_REC, INT)
SGS_IMPL(irssi_textui_textbuffer, cur_line, TEXT_BUFFER_REC, TOBJ(Irssi::TextUI::Line))
SGS_IMPL(irssi_textui_textbuffer, last_eol, TEXT_BUFFER_REC, BOOL)
PROP_SPEC_TBL(irssi_textui_textbuffer) = {
  PROP_GETTER_DEF(irssi_textui_textbuffer, first_line),
  PROP_GETTER_DEF(irssi_textui_textbuffer, lines_count),
  PROP_GETTER_DEF(irssi_textui_textbuffer, cur_line),
  PROP_GETTER_DEF(irssi_textui_textbuffer, last_eol),
  { NULL }
};

/****** Irssi::TextUI::TextBufferView ******/
MAKE_ABSTRACT_CLASS(irssi_textui_textbufferview, Irssi::TextUI::TextBufferView, AC_PRV_STORAGE);
FUN_SPEC_TBL(irssi_textui_textbufferview_proto) = {
  FUN_DEF(irssi_textui_textbufferview, clear, 0), /* Irssi::TextUI::TextBufferView $view */
  FUN_DEF(irssi_textui_textbufferview, get_bookmark, 1), /* Irssi::TextUI::TextBufferView $view, string $name */
  FUN_DEF(irssi_textui_textbufferview, get_line_cache, 1), /* Irssi::TextUI::TextBufferView $view, Irssi::TextUI::Line $line */
  FUN_DEF(irssi_textui_textbufferview, get_lines, 0), /* Irssi::TextUI::TextBufferView $view */
  FUN_DEF(irssi_textui_textbufferview, redraw, 0), /* Irssi::TextUI::TextBufferView $view */
  FUN_DEF(irssi_textui_textbufferview, remove_all_lines, 0), /* Irssi::TextUI::TextBufferView $view */
  FUN_DEF(irssi_textui_textbufferview, remove_line, 1), /* Irssi::TextUI::TextBufferView $view, Irssi::TextUI::Line $line */
  FUN_DEF(irssi_textui_textbufferview, scroll, 1), /* Irssi::TextUI::TextBufferView $view, int $lines */
  FUN_DEF(irssi_textui_textbufferview, scroll_line, 1), /* Irssi::TextUI::TextBufferView $view, Irssi::TextUI::Line $line */
  FUN_DEF(irssi_textui_textbufferview, set_bookmark, 2), /* Irssi::TextUI::TextBufferView $view, string $name, Irssi::TextUI::Line $line */
  FUN_DEF(irssi_textui_textbufferview, set_bookmark_bottom, 1), /* Irssi::TextUI::TextBufferView $view, string $name */
  FUN_DEF(irssi_textui_textbufferview, set_default_indent, 2), /* Irssi::TextUI::TextBufferView $view, int $default_indent, int $longword_noindent */
  FUN_DEF(irssi_textui_textbufferview, set_scroll, 1), /* Irssi::TextUI::TextBufferView $view, int $scroll */
  { NULL }
};
SGS_IMPL(irssi_textui_textbufferview, buffer, TEXT_BUFFER_VIEW_REC, TOBJ(Irssi::TextUI::TextBuffer))
SGS_IMPL(irssi_textui_textbufferview, width, TEXT_BUFFER_VIEW_REC, INT)
SGS_IMPL(irssi_textui_textbufferview, height, TEXT_BUFFER_VIEW_REC, INT)

SGS_IMPL(irssi_textui_textbufferview, default_indent, TEXT_BUFFER_VIEW_REC, INT)
SGS_IMPL(irssi_textui_textbufferview, longword_noindent, TEXT_BUFFER_VIEW_REC, BOOL)
SGS_IMPL(irssi_textui_textbufferview, scroll, TEXT_BUFFER_VIEW_REC, BOOL)

SGS_IMPL(irssi_textui_textbufferview, ypos, TEXT_BUFFER_VIEW_REC, INT)

SGS_IMPL(irssi_textui_textbufferview, startline, TEXT_BUFFER_VIEW_REC, TOBJ(Irssi::TextUI::Line))
SGS_IMPL(irssi_textui_textbufferview, subline, TEXT_BUFFER_VIEW_REC, INT)

SGS_IMPL(irssi_textui_textbufferview, bottom_startline, TEXT_BUFFER_VIEW_REC, TOBJ(Irssi::TextUI::Line))
SGS_IMPL(irssi_textui_textbufferview, bottom_subline, TEXT_BUFFER_VIEW_REC, INT)

SGS_IMPL(irssi_textui_textbufferview, empty_linecount, TEXT_BUFFER_VIEW_REC, INT)
SGS_IMPL(irssi_textui_textbufferview, bottom, TEXT_BUFFER_VIEW_REC, BOOL)
PROP_SPEC_TBL(irssi_textui_textbufferview) = {
  PROP_GETTER_DEF(irssi_textui_textbufferview, buffer),
  PROP_GETTER_DEF(irssi_textui_textbufferview, width),
  PROP_GETTER_DEF(irssi_textui_textbufferview, height),
  PROP_GETTER_DEF(irssi_textui_textbufferview, default_indent),
  PROP_GETTER_DEF(irssi_textui_textbufferview, longword_noindent),
  PROP_GETTER_DEF(irssi_textui_textbufferview, scroll),
  PROP_GETTER_DEF(irssi_textui_textbufferview, ypos),
  PROP_GETTER_DEF(irssi_textui_textbufferview, startline),
  PROP_GETTER_DEF(irssi_textui_textbufferview, subline),
  PROP_GETTER_DEF(irssi_textui_textbufferview, bottom_startline),
  PROP_GETTER_DEF(irssi_textui_textbufferview, bottom_subline),
  PROP_GETTER_DEF(irssi_textui_textbufferview, empty_linecount),
  PROP_GETTER_DEF(irssi_textui_textbufferview, bottom),
  { NULL }
};

/****** Irssi::TextUI::StatusbarItem ******/
MAKE_ABSTRACT_CLASS(irssi_textui_statusbaritem, Irssi::TextUI::StatusbarItem, AC_PRV_STORAGE);
FUN_SPEC_TBL(irssi_textui_statusbaritem_proto) = {
  FUN_DEF(irssi_textui_statusbaritem, default_handler, 4),
  { NULL }
};
SGS_IMPL(irssi_textui_statusbaritem, min_size, SBAR_ITEM_REC, INT)
SGS_IMPL(irssi_textui_statusbaritem, max_size, SBAR_ITEM_REC, INT)
SGS_IMPL(irssi_textui_statusbaritem, xpos, SBAR_ITEM_REC, INT)
SGS_IMPL(irssi_textui_statusbaritem, size, SBAR_ITEM_REC, INT)
SG_IMPL(irssi_textui_statusbaritem, window, SBAR_ITEM_REC, TOBJ(Irssi::UI::Window), N->bar->parent_window ? N->bar->parent_window->active : NULL)

PROP_SPEC_TBL(irssi_textui_statusbaritem) = {
  PROP_GETTER_DEF(irssi_textui_statusbaritem, min_size),
  PROP_GETTER_DEF(irssi_textui_statusbaritem, max_size),
  PROP_GETTER_DEF(irssi_textui_statusbaritem, xpos),
  PROP_GETTER_DEF(irssi_textui_statusbaritem, size),
  PROP_GETTER_DEF(irssi_textui_statusbaritem, window),
  { NULL }
};

/****** Irssi::UI::Theme ******/
MAKE_ABSTRACT_CLASS(irssi_ui_theme, Irssi::UI::Theme, AC_PRV_STORAGE);
FUN_SPEC_TBL(irssi_ui_theme_proto) = {
  FUN_DEF(irssi_ui_theme, format_expand, 2), /* Irssi::UI::Theme $theme, string $format, int $flags */
  FUN_DEF(irssi_ui_theme, get_format, 2), /* Irssi::UI::Theme $theme, string $module, string $tag */
  { NULL }
};
SGS_IMPL(irssi_ui_theme, name, THEME_REC, STR)
PROP_SPEC_TBL(irssi_ui_theme) = {
  PROP_GETTER_DEF(irssi_ui_theme, name),
  { NULL }
};

/****** Irssi::UI::Process ******/
MAKE_ABSTRACT_CLASS(irssi_ui_process, Irssi::UI::Process, AC_PRV_STORAGE);
FUN_SPEC_TBL(irssi_ui_process_proto) = {
  { NULL }
};
SGS_IMPL(irssi_ui_process, id, PROCESS_REC, INT)
SGS_IMPL(irssi_ui_process, name, PROCESS_REC, STR)
SGS_IMPL(irssi_ui_process, args, PROCESS_REC, STR)

SGS_IMPL(irssi_ui_process, pid, PROCESS_REC, INT)
SGS_IMPL(irssi_ui_process, target, PROCESS_REC, STR)
SGS_IMPL(irssi_ui_process, target_win, PROCESS_REC, TOBJ(Irssi::UI::Window))

SGS_IMPL(irssi_ui_process, shell, PROCESS_REC, INT)
SGS_IMPL(irssi_ui_process, notice, PROCESS_REC, INT)
SGS_IMPL(irssi_ui_process, silent, PROCESS_REC, INT)
PROP_SPEC_TBL(irssi_ui_process) = {
  PROP_GETTER_DEF(irssi_ui_process, id),
  PROP_GETTER_DEF(irssi_ui_process, name),
  PROP_GETTER_DEF(irssi_ui_process, args),

  PROP_GETTER_DEF(irssi_ui_process, pid),
  PROP_GETTER_DEF(irssi_ui_process, target),
  PROP_GETTER_DEF(irssi_ui_process, target_win),

  PROP_GETTER_DEF(irssi_ui_process, shell),
  PROP_GETTER_DEF(irssi_ui_process, notice),
  PROP_GETTER_DEF(irssi_ui_process, silent),
  { NULL }
};

#if 0
SCRIPT_FUN[Irssi::Ignore :: add_rec] /* Irssi::Ignore $rec */
SCRIPT_FUN[Irssi::Ignore :: update_rec] /* Irssi::Ignore $rec */
#endif

DLLEXPORT
void
gjs_irssi_generic_finalizer(JSFreeOp *fop, JSObject *obj) {
  JSClass *klass;
  if ((klass = JS_GetClass(obj)) && klass->flags & JSCLASS_HAS_PRIVATE) {
    void *native;
    if ((native = JS_GetPrivate(obj))) {
      g_hash_table_remove(irssi_gjs_object_map, native);
    }
  }
}


JSBool
init_classes(JSContext *cx, JSObject *global) {
  nRV;
  jsval tempval;
  
  JS_BeginRequest(cx);
  JS_DefineFunctions(cx, global, gjs_core_funs);
  if (!JS_InitClass(cx, global, NULL, &gjs_irssi_class,
		    gjs_irssi_constructor, 0,
		    gjs_irssi_proto_props, gjs_irssi_proto_funs,
		    NULL, NULL))
    goto out;
  if (!JS_GetProperty(cx, global, "Irssi", &tempval))
    goto out;
  if (!JS_GetProperty(cx, JSVAL_TO_OBJECT(tempval), "prototype", &tempval))
    goto out;
  gjs_irssi_cto = JSVAL_TO_OBJECT(tempval);
  INIT_CLASS_D(value_reference, irssi, "", "ValueReference", NULL, NULL);
  {
    int code;
    gjs_context_eval(js_context,
		     "Object.defineProperty(Irssi.prototype.ValueReference.prototype, "
		     "'valueOf', {"
		     "  enumerable: false,"
		     "  configurable: true,"
		     "  writable: true,"
		     "  value: function(){return this.value.valueOf();}"
		     "});"
		     "(function(){"
		     "let _toSource = Object.prototype.toSource;"
		     "Object.prototype.toSource = function(){"
		     "return _toSource.call(this).replace(/{/, '/*'+this.constructor.name+'*/{');"
		     "};"
		     "Object.prototype._toSource = _toSource;"
		     "})();",
		     -1, "<internal>", &code, NULL);
  }
  INIT_CLASS_S(irssi, ui, "Irssi", "UI");
  INIT_CLASS_PROPS_R(irssi, rawlog, "Irssi", "Rawlog");
  INIT_CLASS_PROPS_R(irssi, log, "Irssi", "Log");
  INIT_CLASS_PROPS_R(irssi, logitem, "Irssi", "Logitem");
  INIT_CLASS_PROPS_R(irssi, chatnet, "Irssi", "Chatnet");
  INIT_CLASS_PROPS_R(irssi, connect, "Irssi", "Connect");
  INIT_CLASS_PROPS_R(irssi, command, "Irssi", "Command");
  INIT_CLASS_PROPS_R(irssi, ignore, "Irssi", "Ignore");
  INIT_CLASS_PROPS_R(irssi, server, "Irssi", "Server");
  INIT_CLASS_PROPS_R(irssi, windowitem, "Irssi", "Windowitem");
  INIT_CLASS_PROPS_R(irssi, nick, "Irssi", "Nick");
  INIT_CLASS2_PROPS_R(irssi, channel, "Irssi", "Channel", Irssi::Windowitem);
  INIT_CLASS2_PROPS_R(irssi, query, "Irssi", "Query", Irssi::Windowitem);
  INIT_CLASS_S_PROPS_R(irssi_ui, window, "Irssi_UI", "Window");
  INIT_CLASS_PROPS_R(irssi_ui, textdest, "Irssi_UI", "TextDest");
  INIT_CLASS_PROPS_R(irssi_ui, theme, "Irssi_UI", "Theme");
  INIT_CLASS_PROPS_R(irssi_ui, process, "Irssi_UI", "Process");
  
  INIT_CLASS(irssi, textui, "Irssi", "TextUI");
  INIT_CLASS_PROPS_R(irssi_textui, line, "Irssi_TextUI", "Line");
  INIT_CLASS_PROPS_R(irssi_textui, lineinfo, "Irssi_TextUI", "LineInfo");
  INIT_CLASS_PROPS_R(irssi_textui, textbuffer, "Irssi_TextUI", "TextBuffer");
  INIT_CLASS_PROPS_R(irssi_textui, textbufferview, "Irssi_TextUI", "TextBufferView");
  INIT_CLASS_PROPS_R(irssi_textui, statusbaritem, "Irssi_TextUI", "StatusbarItem");
  rv = JS_TRUE;
 out:
  JS_EndRequest(cx);
  return rv;
}
