#include <string.h>
#include <stdlib.h>
#include <string.h>
#include <locale.h>

#include "console.h"

#include "classes.h"
#include "functions.h"
#include "signals.h"
#include "utils.h"

#include "bindings.h"

GjsContext *js_context DLLEXPORT;

JSObject *gjs_irssi_script_loc DLLEXPORT;
GHashTable *gjs_irssi_iobject_class_map DLLEXPORT;
GHashTable *gjs_irssi_class_table DLLEXPORT;

GHashTable *irssi_gjs_object_map;
static char **include_path = NULL;
static const char *module_identifier __attribute__((__unused__)) = MODULE_NAME;
static MODULE_FILE_REC* irssi_module;
static GModule * proxy;

void
c_bless_recursive_add_props(JSContext *cx, JSObject *p, IrssiGjsClassInfo *ci);

void
c_bless_recursive_add_props(JSContext *cx, JSObject *p, IrssiGjsClassInfo *ci) {
  if (!ci) return;

  if (ci->parent)
    c_bless_recursive_add_props(cx, p,
				g_hash_table_lookup(gjs_irssi_class_table, ci->parent));

  if (ci->ps)
    JS_DefineProperties(cx, p, ci->ps);
}

DLLEXPORT
JSObject *
gjs_irssi_c_bless_object (JSContext *cx, const char *class_name, const void *native) {
  JSObject *p = NULL, *proto = NULL;
  JSClass *jsc = NULL;
  IrssiGjsClassInfo *ci;

  if (!native) return NULL;

  if ((p = g_hash_table_lookup(irssi_gjs_object_map, native)))
    return p;
  
  if (class_name[0] == 'i' /* "iobject" */) {
    const IOBJECT_HEADER *iobj = native;
    if ((ci = g_hash_table_lookup(gjs_irssi_iobject_class_map, GINT_TO_POINTER(iobj->type | (iobj->chat_type << 16))))) {}
    else if ((ci = g_hash_table_lookup(gjs_irssi_iobject_class_map, GINT_TO_POINTER(iobj->type)))) {}
  }
  else if (class_name[0] == 's' /* "siobject" */) {
    const IOBJECT_HEADER *iobj = native;
    if ((ci = g_hash_table_lookup(gjs_irssi_iobject_class_map, GINT_TO_POINTER(iobj->type)))) {}
  }
  else if ((ci = g_hash_table_lookup(gjs_irssi_class_table, class_name))) {}

  if (ci) {
    jsc = ci->klass;
    proto = ci->proto;
  }
  p = JS_NewObject(cx, jsc, proto, NULL);
  if (jsc) {
    JS_SetPrivate(p, (void *)native);
    g_hash_table_insert(irssi_gjs_object_map, (void *)native, p);
  }
  else {
    JS_DefineProperty(cx, p, "unknown_class", safe_make_js_string(cx, class_name), NULL, NULL, FIXED_PROP_FLAGS);
  }
  c_bless_recursive_add_props(cx, p, ci);
  return p;
}

void irssi_object_destroyed_handler(const void *p1) {
  JSObject *p;
  if ((p = g_hash_table_lookup(irssi_gjs_object_map, p1))) {
    nCTX;
    /* JS_TransplantObject(cx, p, JS_NewDeadProxyObject(cx, gjs_get_import_global(cx))); */
    JS_SetPrivate(p, NULL);
    JS_DefineProperty(cx, p, "DEAD", JSVAL_TRUE, NULL, NULL, FIXED_PROP_FLAGS);
  }
  g_hash_table_remove(irssi_gjs_object_map, p1);
}


JSBool
gjs_js_define_irssi_stuff(JSContext  *context,
                          JSObject  *module) {
    JSObject *irssi = JS_NewObject(context, &gjs_irssi_class, NULL, NULL);
    JS_DefineProperties(context, irssi, gjs_irssi_props);
    JS_DefineProperty(context, module, "irssi", OBJECT_TO_JSVAL(irssi), NULL, NULL, GJS_MODULE_PROP_FLAGS);
    return JS_TRUE;
}

void
cmd_js(char *data, void *server, WI_ITEM_REC *item)
{
    if(!*data) return;
    command_runsub("js", data, server, item);
}

void
cmd_js_exec(char *data, void *server, WI_ITEM_REC *item)
{
    jsval rval;
    char *st;
    nCTX;
    uSCX;
    JSObject *script_obj;

    JS_BeginRequest(cx);
    script_obj = JS_NewObject(cx, NULL, NULL, NULL);
    StoSCX(STRING_TO_JSVAL(JS_NewStringCopyZ(cx, "<cmdline>")));
    
    if (!JS_EvaluateScript(cx, script_obj, data, strlen(data), "<cmdline>", 0, &rval))
        rval = JSVAL_VOID;

    RstSCX;
    if(!JSVAL_IS_VOID(rval)){
        printtext_window(NULL, 0, "%s", st = decode_js_string(cx, JS_ValueToString(cx, rval)));
        g_free(st);
    }
    JS_EndRequest(cx);
    gjs_maybe_gc(cx);
}


void
cmd_js_unload(char *data, void *server, WI_ITEM_REC *item);
void
cmd_js_load(char *data, void *server, WI_ITEM_REC *item) {
  char *fname;
  jsval rval;
  nCTX;
  uSCX;
    
  JSObject *script_base;
  JSScript *script_obj;
    
  fname = g_strdup_printf("%s/js/%s.js", get_irssi_dir(), data);
  JS_BeginRequest(cx);
  script_base = JS_NewObject(cx, NULL, NULL, NULL);
  script_obj = JS_CompileUTF8File(cx, script_base, fname);
  g_free(fname);
  if (script_obj != NULL) {
    jsval base = OBJECT_TO_JSVAL(script_base);
    cmd_js_unload(data, server, item);

    {
      jsval script_name = safe_make_js_string(cx, data);
      char *script_cn = decode_js_string(cx, JSVAL_TO_STRING(script_name));
      JS_DefineProperty(cx, gjs_irssi_script_loc, script_cn, base,
			NULL, NULL, FIXED_PROP_FLAGS);
      g_free(script_cn);
      StoSCX(script_name);

      JS_ExecuteScript(cx, script_base, script_obj, &rval);
    
      RstSCX;
    }
  }
  JS_EndRequest(cx);
  gjs_maybe_gc(cx);
}

void
cmd_js_unload(char *data, void *server, WI_ITEM_REC *item) {
  nCTX;

  JS_BeginRequest(cx);
  {
    jsval script_name = safe_make_js_string(cx, data);
    char *script_cn = decode_js_string(cx, JSVAL_TO_STRING(script_name));
    script_signal_remove_all(cx, script_cn);
    JS_DefineProperty(cx, gjs_irssi_script_loc, script_cn, JSVAL_VOID, NULL, NULL, 0);
    JS_DeleteProperty(cx, gjs_irssi_script_loc, script_cn);
    g_free(script_cn);
  }
  if (JS_IsExceptionPending(cx))
    JS_ClearPendingException(cx);
  JS_EndRequest(cx);
  gjs_maybe_gc(cx);
}

JSBool
gjs_irssi_extended_formatter(JSContext *cx, const char *format, JSBool fromJS,
			     jsval **vpp, va_list *app) {
  switch (format[1]) {
  case 'S':
    *va_arg(*app, JSString**) = (JSVAL_IS_VOID(**vpp) || JSVAL_IS_NULL(**vpp))
      ? NULL
      : JS_ValueToString(cx, **vpp);
    ++*vpp;
    break;
  }
  return JS_TRUE;
}


void js_gjs_deinit();

DLLEXPORT
void
js_core_init() {
  const char *program_name;
  JSContext *cx;
  JSObject *global;
  int code;
  int chat_type = 0;
    
  proxy = g_module_open("libjs_proxy.so", 0);
  if (!proxy)
    g_error("JS proxy not loaded");
  g_module_make_resident(proxy);
  program_name = "irssi";
#if !GLIB_CHECK_VERSION(2, 35, 1)
  g_type_init ();
#endif    
  js_context = (GjsContext*) g_object_new(GJS_TYPE_CONTEXT,
					  "search-path", include_path,
					  "program-name", program_name,
					  NULL);
  if (!js_context)
    g_error("Reloading JS not supported");

  cx = gjs_context_get_native_context(js_context);
  JS_SetOptions(cx, (JS_GetOptions(cx)
		     &~JSOPTION_VAROBJFIX
		     &~JSOPTION_DONT_REPORT_UNCAUGHT)
		| JSOPTION_STRICT_MODE);
  global = gjs_get_import_global(cx);
  JS_AddArgumentFormatter(cx, ".S", gjs_irssi_extended_formatter);
  if (!gjs_is_registered_native_module(cx, global, program_name)) {
    gjs_register_native_module(program_name, gjs_js_define_irssi_stuff,
			       GJS_NATIVE_SUPPLIES_MODULE_OBJ);
    gjs_context_eval(js_context, "imports.gi", -1, "<internal>", &code, NULL);
  }
  JS_BeginRequest(cx);
  gjs_irssi_script_loc = JS_DefineObject(cx, global, "Script", NULL, NULL, FIXED_PROP_FLAGS);
  JS_DefineObject(cx, global, "Global", NULL, NULL, FIXED_PROP_FLAGS);

  gjs_irssi_class_table = g_hash_table_new_full(g_str_hash, g_str_equal, NULL, g_free);
  irssi_gjs_object_map = g_hash_table_new(g_direct_hash, NULL);
  gjs_irssi_iobject_class_map = g_hash_table_new(g_direct_hash, NULL);

  if (!init_classes(cx, global))
    goto error_cleanup;

  /* hash = GINT_TO_POINTER(type | (chat_type << 16)); */
  
  CLASS_ADD_MAP(id_str("WINDOW ITEM TYPE", "CHANNEL"), Irssi::Channel);
  CLASS_ADD_MAP(id_str("WINDOW ITEM TYPE", "QUERY"), Irssi::Query);
  CLASS_ADD_MAP(id("NICK", 0), Irssi::Nick);
  CLASS_ADD_MAP(id("CHATNET", 0), Irssi::Chatnet);
  CLASS_ADD_MAP(id("SERVER", 0), Irssi::Server);
  CLASS_ADD_MAP(id("SERVER CONNECT", 0), Irssi::Connect);
    
  if (!init_signals(cx, global))
    goto error_cleanup;

  irssi_module = module_register("js", "core");
  signal_add_last("window destroyed", irssi_object_destroyed_handler);
  signal_add_last("query destroyed", irssi_object_destroyed_handler);
  signal_add_last("channel destroyed", irssi_object_destroyed_handler);
  signal_add_last("chatnet destroyed", irssi_object_destroyed_handler);
  /* ??? XXX */
  signal_add_last("ignore destroyed", irssi_object_destroyed_handler);
  signal_add_last("dcc destroyed", irssi_object_destroyed_handler);
  signal_add_last("keyinfo destroyed", irssi_object_destroyed_handler);
  signal_add_last("theme destroyed", irssi_object_destroyed_handler);
  signal_add_last("mainwindow destroyed", irssi_object_destroyed_handler);
  command_bind("js", NULL, (SIGNAL_FUNC)cmd_js);
  command_bind("js exec", NULL, (SIGNAL_FUNC)cmd_js_exec);
  command_bind("js load", NULL, (SIGNAL_FUNC)cmd_js_load);
  command_bind("js unload", NULL, (SIGNAL_FUNC)cmd_js_unload);

  goto out;
    
error_cleanup:
  js_gjs_deinit();
    
out:
  JS_EndRequest(cx);
}

void js_gjs_deinit() {
  g_object_run_dispose(G_OBJECT(js_context));
  g_object_unref(js_context);
  g_hash_table_destroy(gjs_irssi_iobject_class_map);
  g_hash_table_destroy(gjs_irssi_class_table);
  g_hash_table_destroy(irssi_gjs_object_map);
  g_module_close(proxy);
}

DLLEXPORT
void
js_core_deinit() {
  command_unbind("js", (SIGNAL_FUNC)cmd_js);
  command_unbind("js exec", (SIGNAL_FUNC)cmd_js_exec);
  command_unbind("js load", (SIGNAL_FUNC)cmd_js_load);
  command_unbind("js unload", (SIGNAL_FUNC)cmd_js_unload);
  signal_remove("window destroyed", irssi_object_destroyed_handler);
  js_gjs_deinit();
}

