#ifndef IRSSI_GJS_XS_MACROS_H_
#define IRSSI_GJS_XS_MACROS_H_

/* flags for properties created by native code */
#define FIXED_PROP_FLAGS GJS_MODULE_PROP_FLAGS | JSPROP_READONLY | JSPROP_PERMANENT
#define HID_FIX_PROP_FLAGS JSPROP_READONLY | JSPROP_PERMANENT

/** global variables:
js_context = gjs context
gjs_irssi_script_loc = parent object of script objects
gjs_irssi_script_loc.CURRENT_SCRIPT = script context (name of current script)
gjs_irssi_script_loc.Signals = bound signals object
*/

/** function global variable names:
cx  = javascript context
vp  = arg pointer
sCX = script context (name)
argv
rv
last_script_cx
*/

/** function cleanup labels:
SCX_cleanup:  after failure to get sCX
sObj_cleanup: after failure to get script object
 */

/* js native function signature */
#define SCRIPT_FUN(name) JSBool gjs_ ## name(JSContext *cx, unsigned argc, jsval *vp)
/** variable declaration macros */
/* require argv */
#define nArgv                                   \
    jsval *argv = JS_ARGV(cx, vp)    
/* require script context (string) */
#define nSCX                                    \
    jsval sCX
/* require return value */
#define nRV \
    JSBool rv = JS_FALSE
/* require previous script ctx for save+restore */
#define uSCX                                    \
    jsval last_script_cx
/* get context from global static variable */
#define nCTX                                                    \
    JSContext *cx = gjs_context_get_native_context(js_context)

/** variable modification macros */
/* save sctx and set it to ARG */
#define StoSCX(script_cx)                             \
  if (!JS_GetProperty(cx, gjs_irssi_script_loc, "CURRENT_SCRIPT", &last_script_cx)) \
    last_script_cx = JSVAL_VOID;					\
  JS_DefineProperty(cx, gjs_irssi_script_loc, "CURRENT_SCRIPT", script_cx,        \
                      NULL, NULL, FIXED_PROP_FLAGS)
/* save sctx and set it to sCX ("current" script) */
#define StocSCX StoSCX(sCX)
/* restore sctx after Sto*SCX */
#define RstSCX \
    JS_DefineProperty(cx, gjs_irssi_script_loc, "CURRENT_SCRIPT", last_script_cx, \
                      NULL, NULL, 0)

/* get script context from jsobject ARG1 property ARG2 */
#define get_pSCX(loc, var) do{                                          \
  if (!JS_GetProperty(cx, loc, var, &sCX))				\
    goto SCX_cleanup;							\
  if (!JSVAL_IS_STRING(sCX)) {						\
    JS_ReportError(cx, "No Script context found");			\
    goto SCX_cleanup;							\
  } }while(0)
/* get "current" script from global variable */
#define get_cSCX get_pSCX(gjs_irssi_script_loc, "CURRENT_SCRIPT")
/* get script object of sCX into ARG */
#define get_sObj(obj) do{                       \
    char *script_name;							\
    JSBool succ;							\
    succ = JS_GetProperty(cx, gjs_irssi_script_loc,                                      \
			  script_name = JS_EncodeString(cx, JSVAL_TO_STRING(sCX)), &obj); \
    JS_free(cx, script_name);                                           \
    if (!succ)								\
      goto sObj_cleanup;						\
    if (JSVAL_IS_PRIMITIVE(obj)) {					\
      g_message("No script object");					\
      goto sObj_cleanup;						\
    } } while(0)

#define AC_NONE 0
#define AC_PRV_STORAGE JSCLASS_HAS_PRIVATE

/** class factory:
 * MAKE_ABSTRACT_CLASS
 * \param klass class name with underscores
 * \param class_name string class name
 * \param flags implementation-specific flags
 */

#define DECLARE_CLASS(klass)			\
  JSObject *gjs_ ## klass ## _cto, *gjs_ ## klass ## _proto;	\
  extern JSClass gjs_ ## klass ## _class
#define MAKE_ABSTRACT_CLASS(klass, class_name, flags)	\
  GJS_NATIVE_CONSTRUCTOR_DEFINE_ABSTRACT(klass)		\
JSClass gjs_ ## klass ## _class = { #class_name,	\
    flags, \
    JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_StrictPropertyStub, \
    JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, gjs_irssi_generic_finalizer, \
    JSCLASS_NO_OPTIONAL_MEMBERS \
}

#define SCRIPT_GETTER(klass, prop) \
JSBool                                                            \
static gjs_ ## ## klass ## _get_ ## prop                       \
 (JSContext *cx, JSHandleObject obj, JSHandleId id, JSMutableHandleValue vp)

/** getter entry in implementation object table
 * SIMPLE_GETTER_DEF
 * \param klass class name with underscores
 * \param prop attribute name with underscores
 */
#define PROP_SPEC_TBL_NAME(klass)		\
  gjs_ ## klass ## _props
#define PROP_SPEC_TBL(klass)		\
  JSPropertySpec PROP_SPEC_TBL_NAME(klass)[]
#define DECLARE_PROP_SPEC_TBL(klass)		\
  extern PROP_SPEC_TBL(klass)
  
#define PROP_GETTER_DEF(klass, prop) \
    { #prop, 0,                                                \
    GJS_MODULE_PROP_FLAGS | JSPROP_READONLY,                      \
    JSOP_WRAPPER(gjs_ ## klass ## _get_ ## prop),      \
    JSOP_WRAPPER(JS_StrictPropertyStub)                           \
    }

#define FUN_SPEC_TBL_NAME(klass)		\
  gjs_ ## klass ## _funs
#define FUN_SPEC_TBL(klass)		\
  JSFunctionSpec FUN_SPEC_TBL_NAME(klass)[]

#define FUN_DEF(klass, prop, argc)					\
  JS_FS(#prop, gjs_ ## klass ## _ ## prop, argc, GJS_MODULE_PROP_FLAGS)


#define CLASS_ADD_MAP(x,y)							\
  g_hash_table_insert(gjs_irssi_iobject_class_map,			\
		      GINT_TO_POINTER( (chat_type << 16)		\
				       | module_get_uniq_ ## x ),	\
		      g_hash_table_lookup(gjs_irssi_class_table, #y ))

#endif
